@@ -1,6 +1,6 @@
 import warnings
 from fault.verilog_target import VerilogTarget
-from .verilog_utils import verilog_name
+from .verilog_utils import verilog_name, is_nd_array
 from .util import (is_valid_file_mode, file_mode_allows_reading,
                    file_mode_allows_writing)
 import magma as m
@@ -53,14 +53,15 @@ class SystemVerilogTarget(VerilogTarget):
                  skip_compile=None, magma_output="coreir-verilog",
                  magma_opts=None, include_verilog_libraries=None,
                  simulator=None, timescale="1ns/1ns", clock_step_delay=5,
-                 num_cycles=10000, dump_waveforms=True, dump_vcd=None,
-                 no_warning=False, sim_env=None, ext_model_file=None,
-                 ext_libs=None, defines=None, flags=None, inc_dirs=None,
-                 ext_test_bench=False, top_module=None, ext_srcs=None,
-                 use_input_wires=False, parameters=None, disp_type='on_error',
-                 waveform_file=None, coverage=False, use_kratos=False,
-                 use_sva=False, skip_run=False, no_top_module=False,
-                 vivado_use_system_verilog=True):
+                 num_cycles=10000, dump_waveforms=False, dump_vcd=None,
+                 waveform_type=None, no_warning=False, sim_env=None,
+                 ext_model_file=None, ext_libs=None, defines=None, flags=None,
+                 inc_dirs=None, ext_test_bench=False, top_module=None,
+                 ext_srcs=None, use_input_wires=False, parameters=None,
+                 disp_type='on_error', waveform_file=None, coverage=False,
+                 use_kratos=False, use_sva=False, skip_run=False,
+                 no_top_module=False, vivado_use_system_verilog=True,
+                 disable_ndarray=False):
         """
         circuit: a magma circuit
 
@@ -135,6 +136,9 @@ class SystemVerilogTarget(VerilogTarget):
 
         dump_waveforms: Enable tracing of internal values
 
+        waveform_type: 'vcd', 'vpd', 'fsdb'.  Default for ncsim is 'vcd'. Default
+                       for vcs is 'vpd'.
+
         waveform_file: name of file to dump waveforms (default is
                        "waveform.vcd" for ncsim and "waveform.vpd" for vcs)
 
@@ -149,6 +153,13 @@ class SystemVerilogTarget(VerilogTarget):
         vivado_use_system_verilog: If True (default), mark Vivado source files
                                    as SystemVerilog so that more modern syntax
                                    is supported.
+
+        disable_ndarray: If True, disable magma/fault support for
+                         codegenerating verilog ndarrays (multi-dimensional
+                         arrays of bits).
+                         Default is False except when the simulator is
+                         "iverilog" when it is always True (since iverilog does
+                         not currently support unpacked arrays)
         """
         # set default for list of external sources
         if include_verilog_libraries is None:
@@ -168,6 +179,12 @@ class SystemVerilogTarget(VerilogTarget):
         # set default for magma compilation options
         magma_opts = magma_opts if magma_opts is not None else {}
 
+        if simulator == "iverilog":
+            disable_ndarray = True
+        if disable_ndarray:
+            magma_opts["disable_ndarray"] = True
+        self.disable_ndarray = disable_ndarray
+
         # call the super constructor
         super().__init__(circuit, circuit_name, directory, skip_compile,
                          include_verilog_libraries, magma_output,
@@ -219,9 +236,17 @@ class SystemVerilogTarget(VerilogTarget):
         self.disp_type = disp_type
         self.waveform_file = waveform_file
         self.use_sva = use_sva
+        self.waveform_type = waveform_type
         if self.waveform_file is None and self.dump_waveforms:
             if self.simulator == "vcs":
-                self.waveform_file = "waveforms.vpd"
+                if self.waveform_type is None:
+                    suffix = "vpd"
+                else:
+                    if self.waveform_type not in ["vpd", "fsdb"]:
+                        raise ValueError("Only 'vpd' and 'fsdb' supported for "
+                                         "vcs waveform type")
+                    suffix = self.waveform_type
+                self.waveform_file = f"waveforms.{self.waveform_type}"
             elif self.simulator in {"ncsim", "iverilog", "vivado"}:
                 self.waveform_file = "waveforms.vcd"
             else:
@@ -262,21 +287,25 @@ class SystemVerilogTarget(VerilogTarget):
             port = port.select_path
         if isinstance(port, SelectPath):
             if len(port) > 2:
-                name = f"dut.{port.system_verilog_path}"
+                name = f"dut.{port.system_verilog_path(self.disable_ndarray)}"
             else:
                 # Top level ports assign to the external reg
-                name = verilog_name(port[-1].name)
+                name = verilog_name(port[-1].name, self.disable_ndarray)
         elif isinstance(port, fault.WrappedVerilogInternalPort):
             name = f"dut.{port.path}"
         else:
-            name = verilog_name(port.name)
+            name = verilog_name(port.name, self.disable_ndarray)
         return name
 
     def process_peek(self, value):
         if isinstance(value.port, fault.WrappedVerilogInternalPort):
             return f"dut.{value.port.path}"
-        else:
-            return f"{value.port.name}"
+        elif isinstance(value.port, PortWrapper):
+            path = value.port.select_path.system_verilog_path(
+                self.disable_ndarray
+            )
+            return f"dut.{path}"
+        return f"{value.port.name}"
 
     def make_var(self, i, action):
         if isinstance(action._type, AbstractBitVectorMeta):
@@ -305,7 +334,8 @@ class SystemVerilogTarget(VerilogTarget):
         elif isinstance(value, actions.Peek):
             value = self.process_peek(value)
         elif isinstance(value, PortWrapper):
-            value = f"dut.{value.select_path.system_verilog_path}"
+            path = value.select_path.system_verilog_path(self.disable_ndarray)
+            value = f"dut.{path}"
         elif isinstance(value, actions.FileRead):
             new_value = f"{value.file.name_without_ext}_in"
             value = new_value
@@ -318,13 +348,20 @@ class SystemVerilogTarget(VerilogTarget):
             left = self.compile_expression(value.left)
             right = self.compile_expression(value.right)
             op = value.op_str
+            if op == "==":
+                # Use strict eq
+                op = "==="
+            elif op == "!=":
+                # Use strict neq
+                op = "!=="
             return f"({left}) {op} ({right})"
         elif isinstance(value, expression.UnaryOp):
             operand = self.compile_expression(value.operand)
             op = value.op_str
             return f"{op} ({operand})"
         elif isinstance(value, PortWrapper):
-            return f"dut.{value.select_path.system_verilog_path}"
+            path = value.select_path.system_verilog_path(self.disable_ndarray)
+            return f"dut.{path}"
         elif isinstance(value, actions.Peek):
             return self.process_peek(value)
         elif isinstance(value, actions.Var):
@@ -514,7 +551,7 @@ class SystemVerilogTarget(VerilogTarget):
         return ['#1;']
 
     def make_step(self, i, action):
-        name = verilog_name(action.clock.name)
+        name = verilog_name(action.clock.name, self.disable_ndarray)
         code = []
         for step in range(action.steps):
             code.append(f"#{self.clock_step_delay} {name} ^= 1;")
@@ -539,7 +576,16 @@ class SystemVerilogTarget(VerilogTarget):
     def generate_port_code(self, name, type_, power_args):
         is_array_of_non_bits = issubclass(type_, m.Array) and \
             not issubclass(type_.T, m.Bit)
-        if is_array_of_non_bits or issubclass(type_, m.Tuple):
+        if is_nd_array(type_) and not self.disable_ndarray:
+            outer_width = ""
+            while not issubclass(type_.T, m.Digital):
+                outer_width += f"[{type_.N - 1}:0]"
+                type_ = type_.T
+            inner_width = f"[{type_.N - 1}:0]"
+            t = "reg" if type_.is_input() else "wire"
+            self.add_decl(f'{t} {inner_width}', f'{name} {outer_width}')
+            return [f".{name}({name})"]
+        elif is_array_of_non_bits or issubclass(type_, m.Tuple):
             return self.generate_recursive_port_code(name, type_, power_args)
         else:
             width_str = ""
@@ -596,9 +642,13 @@ class SystemVerilogTarget(VerilogTarget):
 
         # set up probing
         if self.dump_waveforms and self.simulator == "vcs":
-            initial_body += [f'$vcdplusfile("{self.waveform_file}");',
-                             f'$vcdpluson();',
-                             f'$vcdplusmemon();']
+            if self.waveform_type == "vpd":
+                initial_body += [f'$vcdplusfile("{self.waveform_file}");',
+                                 f'$vcdpluson();',
+                                 f'$vcdplusmemon();']
+            if self.waveform_type == "fsdb":
+                initial_body += [f'$fsdbDumpfile("{self.waveform_file}");',
+                                 f'$fsdbDumpvars();']
         elif self.dump_waveforms and self.simulator in {"iverilog", "vivado"}:
             # https://iverilog.fandom.com/wiki/GTKWAVE
             initial_body += [f'$dumpfile("{self.waveform_file}");',
@@ -692,14 +742,14 @@ class SystemVerilogTarget(VerilogTarget):
             sim_err_str = None
             # Run simulation
             bin_cmd = [bin_file]
-            bin_err_str = 'Error'
+            bin_err_str = ['Error', 'Fatal']
         elif self.simulator == 'iverilog':
             # Compile simulation
             sim_cmd, bin_file = self.iverilog_cmd(sources=vlog_srcs)
             sim_err_str = ['syntax error', 'I give up.']
             # Run simulation
             bin_cmd = ['vvp', '-N', bin_file]
-            bin_err_str = 'ERROR'
+            bin_err_str = ['ERROR', 'FATAL']
         else:
             raise NotImplementedError(self.simulator)
 
@@ -977,7 +1027,10 @@ class SystemVerilogTarget(VerilogTarget):
             raise NotImplementedError("coverage in vcs is not implemented yet")
 
         if self.dump_waveforms:
-            cmd += ['+vcs+vcdpluson', '-debug_pp']
+            cmd += ['-debug_pp']
+
+            if self.waveform_type == "vcd":
+                cmd += ['+vcs+vcdpluson']
 
         # specify top module
         if not self.no_top_module:
@@ -1023,13 +1076,14 @@ class SystemVerilogTarget(VerilogTarget):
         # misc flags
         cmd += ['-g2012']
 
-        # source files
-        cmd += [f'{src}' for src in sources]
-
-        # set the top module
+        # set the top module.  note that the '-s' option must
+        # come before the source file list.
         if not self.no_top_module:
             cmd += ['-s', f'{self.top_module}']
 
+        # source files
+        cmd += [f'{src}' for src in sources]
+
         # return arg list and binary file location
         return cmd, bin_file
 
@@ -1040,7 +1094,7 @@ class SynchronousSystemVerilogTarget(SystemVerilogTarget):
             raise ValueError("Clock required")
 
         super().__init__(*args, **kwargs)
-        name = verilog_name(clock.name)
+        name = verilog_name(clock.name, self.disable_ndarray)
         self.clock_drivers.append(
             f"always #{self.clock_step_delay} {name} = ~{name};"
         )
